/*
 * Copyright (C) 1996-2001 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
/* $LAAS: gdhe_terrain.h,v 1.4 2003/02/09 16:31:26 matthieu Exp $ */

#ifndef _TERRAIN_H
#define _TERRAIN_H

/*
 * structure utilisee pour representer les terrains
 */
/* un point */
typedef struct {
    short z;			/* elevation en unites (dz) */
    unsigned short type;		/* type d'un point */
					/* type == 0 && z == 0 
					   -> pas de donne'es en ce point */
} GDHE_MNT_POINT;

/* indique le type de stockage d'un MNT */
typedef enum {
    GDHE_MNT_X_COLUMN,			/* X varie avec j (colonnes) */
    GDHE_MNT_X_ROW			/* X varie avec i (lignes) */
} GDHE_MNT_STORAGE_TYPE;

/* l'entete */   
typedef struct {
    double x0, y0, z0;			/* coordonne's du point origine */
    double dx, dy, dz;			/* variations en x, y, z/unite */
    unsigned int nrows, ncols;		/* nombre de lignes et colonnes */
    GDHE_MNT_STORAGE_TYPE storageType;	/* type de rangement */
} GDHE_MNT_HEADER;

/* la carte */
typedef struct {
    GDHE_MNT_HEADER hdr;
    GDHE_MNT_POINT *points;
} GDHE_MNT;

#ifdef GDHE_PRIVATE 
/*
 * structures internes a GDHE 
 */
/* sommet pret pour dessin OpenGL */
typedef struct {
    float point[3];
    float normal[3];
    unsigned int type;
} GDHE_MNT_POINT_INT;

/* les donnees d'un terrain */
typedef struct {
    double x0, y0, z0;		/* coordonnes du point origine */
    double dx, dy, dz;		/* variations en x, y, z/unite */
    int nrows, ncols;
    GDHE_MNT_POINT_INT *points;
} GDHE_MNT_INT;

#endif /* GDHE_PRIVATE */
#endif
