/*
 * Copyright (C) 1996-2001 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"
__RCSID("$LAAS: jmem.c,v 1.3 2003/02/09 16:31:28 matthieu Exp $");

#include <sys/types.h>
#include <stdio.h>

#include "jpeglib.h"
#include "jerror.h"

/**
 ** Driver d'endtree/sortie de pour libjpeg pour compression en memoire 
 **/

typedef struct {
    struct jpeg_destination_mgr pub;	/* public fields */
    JOCTET *buffer;
    int size;
} my_destination_mgr;

typedef my_destination_mgr *my_dest_ptr;

typedef struct {
    struct jpeg_source_mgr pub;		/* public fields */
    JOCTET *buffer;
    boolean start_of_file;		/* have we gotten any data yet? */
    int size;
} my_source_mgr;

typedef my_source_mgr *my_src_ptr;

/*
 * Initialize destination
 */
METHODDEF(void)
init_destination(j_compress_ptr cinfo)
{
    my_dest_ptr dest = (my_dest_ptr)cinfo->dest;

    dest->pub.next_output_byte = dest->buffer;
    dest->pub.free_in_buffer = dest->size;
}

/*
 * Empty the output buffer
 */
METHODDEF(boolean)
empty_output_buffer(j_compress_ptr cinfo)
{
    my_dest_ptr dest = (my_dest_ptr)cinfo->dest;
    
    dest->pub.next_output_byte = dest->buffer;
    dest->pub.free_in_buffer = dest->size;
    return TRUE;
}

/*
 * Terminate destination
 */
METHODDEF(void)
term_destination(j_compress_ptr cinfo)
{
    return;
}

/*
 * Prepare for output to a  memory buffer
 */
GLOBAL(void)
jpeg_mem_dest(j_compress_ptr cinfo, unsigned char *buffer, int size)
{
    my_dest_ptr dest;

    if (cinfo->dest == NULL) {
	cinfo->dest = (struct jpeg_destination_mgr *)
		      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 
						  JPOOL_PERMANENT,
						  sizeof(my_destination_mgr));
    }
  dest = (my_dest_ptr) cinfo->dest;
  dest->pub.init_destination = init_destination;
  dest->pub.empty_output_buffer = empty_output_buffer;
  dest->pub.term_destination = term_destination;
  dest->buffer = buffer;
  dest->size = size;
}

GLOBAL(int)
jpeg_mem_size(j_compress_ptr cinfo)
{
    my_dest_ptr dest = (my_dest_ptr)cinfo->dest;
    
    if (dest) {
	return (dest->pub.next_output_byte - dest->buffer);
    } else {
	return -1;
    }
}

/*----------------------------------------------------------------------*/

METHODDEF(void)
init_source(j_decompress_ptr cinfo)
{
    my_src_ptr src = (my_src_ptr)cinfo->src;

    /* We reset the empty-input-file flag for each image,
     * but we don't clear the input buffer.
     * This is correct behavior for reading a series 
     of images from one source.
     */
    src->start_of_file = TRUE;
}

METHODDEF(boolean)
fill_input_buffer(j_decompress_ptr cinfo)
{
    my_src_ptr src = (my_src_ptr)cinfo->src;
    int l;

    src->pub.next_input_byte = src->buffer;
    src->pub.bytes_in_buffer = src->size;
    src->start_of_file = FALSE;
    return TRUE;
}

METHODDEF(void)
skip_input_data(j_decompress_ptr cinfo, long num_bytes)
{
    my_src_ptr src = (my_src_ptr)cinfo->src;
    
    if (num_bytes > 0) {
	while (num_bytes > (long) src->pub.bytes_in_buffer) {
	    num_bytes -= (long) src->pub.bytes_in_buffer;
	    (void) fill_input_buffer(cinfo);
	    /* note we assume that fill_input_buffer will never return FALSE,
	     * so suspension need not be handled.
	     */
	}
	src->pub.next_input_byte += (size_t) num_bytes;
	src->pub.bytes_in_buffer -= (size_t) num_bytes;
    }
}

METHODDEF(void)
    term_source(j_decompress_ptr cinfo)
{
    /* Nothing to do */
}

/*
 * Prepare for inpuf from a  memory buffer
 */
GLOBAL(void)
jpeg_mem_src(j_decompress_ptr cinfo, unsigned char *buffer, int length)
{
    my_src_ptr src;

    if (cinfo->src == NULL) {
	cinfo->src = (struct jpeg_source_mgr *)
		     (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, 
						 JPOOL_PERMANENT,
						 sizeof(my_source_mgr));
    }
    src = (my_src_ptr) cinfo->src;
    src->pub.init_source = init_source;
    src->pub.fill_input_buffer = fill_input_buffer;
    src->pub.skip_input_data = skip_input_data;
    src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
    src->pub.term_source = term_source;
    src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
    src->pub.next_input_byte = NULL; /* until buffer loaded */

    src->buffer = buffer;
    src->size = length;
}


