/*
 * Copyright (C) 1996-2001 LAAS/CNRS 
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 *    - Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *    - Redistributions in binary form must reproduce the above
 *      copyright notice, this list of conditions and the following
 *      disclaimer in the documentation and/or other materials provided
 *      with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
 * ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */
#include "config.h"
__RCSID("$LAAS: tkjpeg.c,v 1.3 2003/02/09 16:31:28 matthieu Exp $");

/***
 *** Handlers jpeg pour TK
 ***/
#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <setjmp.h>
#include <tk.h>
#undef EXTERN
#include "jpeglib.h"
#include "jerror.h"

#include "jmem.h"


#define TCL_JPEG_UNKNOWN 0
#define TCL_JPEG_MONO    1
#define TCL_JPEG_COLOR   2

/* Handler d'erreur */
typedef struct {
    struct jpeg_error_mgr pub;		/* donnees publiques */
    jmp_buf setjmp_buffer;		/* Adresse de retour */
} my_error_mgr, *my_error_ptr;

/* Fonction de declaration des methodes de lecture */
extern void jpeg_tclChannel_src(j_decompress_ptr cinfo, Tcl_Channel infile);

/*----------------------------------------------------------------------*/

/**
 ** Gestionnaire des erreurs fatales pour permette une reprise
 **/
static void
my_error_exit(j_common_ptr cinfo)
{
    my_error_ptr myerr = (my_error_ptr)cinfo->err;
    /* Saut a l'adresse de reprise memorisee */
    longjmp(myerr->setjmp_buffer, 1);
}

/*----------------------------------------------------------------------*/

/***
 ***  Handlers pour widget photo de TK
 ***/

/* Variables globales */
static struct jpeg_decompress_struct cinfo;
static my_error_mgr jerr;

/**
 ** Initialisation de la bibliotheque jpeg
 ** Lecture du header d'une image 
 **/
static int 
fileReadJpegHeader(Tcl_Channel chan, int *widthPtr, int *heightPtr)
{
    int status; 
    void (*error_exit_handler)(j_common_ptr);

    /* Definition des handlers d'erreur */
    cinfo.err = jpeg_std_error((struct jpeg_error_mgr *)&jerr);
    /* Allocation d'une structure de decompression */
    jpeg_create_decompress(&cinfo);
    /* Lien avec le canal TCL */
    jpeg_tclChannel_src(&cinfo, chan);
    /* Sauvegarde du handler d'erreur fatale */
    error_exit_handler = jerr.pub.error_exit;
    /* installation d'un handler local qui fait un lonjmp */
    jerr.pub.error_exit = my_error_exit;
    /* Lecture du header avec protection contre les erreurs fatales */
    if (setjmp(jerr.setjmp_buffer) == 0) {
	status = jpeg_read_header(&cinfo, FALSE);
	if (status != JPEG_HEADER_OK) {
	    return 0;
	}
    } else {
	/* Une erreur fatale s'est produite dans jpeg_read_header */
	jpeg_destroy_decompress(&cinfo);
	return 0;
    }
    /* Restaure le handler d'erreur fatales par defaut */
    jerr.pub.error_exit = error_exit_handler;
    /* Valeurs retourne'es */
    *widthPtr = cinfo.image_width;
    *heightPtr = cinfo.image_height;
    switch (cinfo.num_components) {
      case 1:
	return TCL_JPEG_MONO;
      case 3:
	return TCL_JPEG_COLOR;
      default:
	/* Nombre de composantes non traite' */
	return TCL_JPEG_UNKNOWN;
    }
}

/*----------------------------------------------------------------------*/

static int
dataReadJpegHeader(Tcl_Obj *data, int *widthPtr, int *heightPtr)
{
    int status, length;
    void (*error_exit_handler)(j_common_ptr);
    unsigned char *p;
    
    p = Tcl_GetByteArrayFromObj(data, &length);
    /* Definition des handlers d'erreur */
    cinfo.err = jpeg_std_error((struct jpeg_error_mgr *)&jerr);
    /* Allocation d'une structure de decompression */
    jpeg_create_decompress(&cinfo);
    /* Lien avec le canal TCL */
    jpeg_mem_src(&cinfo, p, length);
    /* Sauvegarde du handler d'erreur fatale */
    error_exit_handler = jerr.pub.error_exit;
    /* installation d'un handler local qui fait un lonjmp */
    jerr.pub.error_exit = my_error_exit;
    /* Lecture du header avec protection contre les erreurs fatales */
    if (setjmp(jerr.setjmp_buffer) == 0) {
	status = jpeg_read_header(&cinfo, FALSE);
	if (status != JPEG_HEADER_OK) {
	    return 0;
	}
    } else {
	/* Une erreur fatale s'est produite dans jpeg_read_header */
	jpeg_destroy_decompress(&cinfo);
	return 0;
    }
    /* Restaure le handler d'erreur fatales par defaut */
    jerr.pub.error_exit = error_exit_handler;
    /* Valeurs retourne'es */
    *widthPtr = cinfo.image_width;
    *heightPtr = cinfo.image_height;
    switch (cinfo.num_components) {
      case 1:
	return TCL_JPEG_MONO;
      case 3:
	return TCL_JPEG_COLOR;
      default:
	/* Nombre de composantes non traite' */
	return TCL_JPEG_UNKNOWN;
    }
}
    
/*----------------------------------------------------------------------*/

/**
 ** Teste si on est en presence d'une image jpeg
 **/
int 
jpegFileMatchProc(Tcl_Channel chan, const char *filename, 
		  Tcl_Obj *formatString,
		  int *widthPtr, int *heightPtr, Tcl_Interp *interp)
{
    int result;
    /* Essaye de lire l'entete */
    result = fileReadJpegHeader(chan, widthPtr, heightPtr);
    /* libere la structure de decompression correspondante */
    jpeg_destroy_decompress(&cinfo);
    return result;
}

/*----------------------------------------------------------------------*/

int
jpegDataMatchProc(Tcl_Obj *data, Tcl_Obj *formatString, 
		  int *widthPtr, int *heightPtr,  Tcl_Interp *interp)
{
    int result;

    result = dataReadJpegHeader(data, widthPtr, heightPtr);
    jpeg_destroy_decompress(&cinfo);
    return result;
}

/*----------------------------------------------------------------------*/

/**
 ** Lecture d'un morceau d'image
 **/
int
jpegFileReadProc(Tcl_Interp *interp, Tcl_Channel chan,
		 const char *fileName, Tcl_Obj *formatString, 
		 Tk_PhotoHandle imageHandle,
		 int destX, int destY, int width, int height,
		 int srcX, int srcY)
{
    int fileWidth, fileHeight;
    int nLines, nBytes, h, type, count;
    unsigned char *pixelPtr;
    unsigned char **rowPtr;
    Tk_PhotoImageBlock block;
    
    /* Lecture de l'entete */
    type = fileReadJpegHeader(chan, &fileWidth, &fileHeight);
    if (type == 0) {
	/* source non reconnue comme jpeg */
	Tcl_AppendResult(interp, 
			 "couldn't read jpeg header from file\"", 
			 fileName, "\"", NULL);
	return TCL_ERROR;
    }
    if ((fileWidth <= 0) || (fileHeight <= 0)) {
	/* Dimensions negatives ou nulles */
	Tcl_AppendResult(interp, "Jpeg file \"", fileName, 
			 "\" has dimension(s) <= 0", NULL);
	return TCL_ERROR;
    }
    /* Calcul de la zone a lire */
    if ((srcX + width) > fileWidth) {
        width = fileWidth - srcX;
    }
    if ((srcY + height) > fileHeight) {
        height = fileHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
        || (srcX >= fileWidth) || (srcY >= fileHeight)) {
        return TCL_OK;
    }
    /* Debut de la decompression */
    jpeg_start_decompress(&cinfo);
    
    if (type == TCL_JPEG_MONO) {
	block.pixelSize = 1;
	block.offset[0] = 0;
	block.offset[1] = 0;
	block.offset[2] = 0;
    } else {
	block.pixelSize = 3;
        block.offset[0] = 0;
	block.offset[1] = 1;
	block.offset[2] = 2;
    }
    block.width = width;
    block.pitch = block.pixelSize * fileWidth;

    /* Allocation structure photo */
    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

#if 0
    /* XXX */
    if (srcY > 0) {
        Tcl_Seek(chan, (srcY * block.pitch), SEEK_CUR);
    }
#endif
    /*  buffer pour  le nombre de lignes prescrit par libjpeg */
    nLines = cinfo.rec_outbuf_height;
    nBytes = nLines * block.pitch;
    pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
    rowPtr = (unsigned char **) ckalloc((unsigned) nLines);
    /* Initialisation des pointeurs sur les lignes */
    for (h = 0; h < nLines; h++) {
	rowPtr[h] = pixelPtr + h*block.pitch;
    }
    block.pixelPtr = pixelPtr + srcX * block.pixelSize; /* XXXX ??? */

    /* Boucle de lecture/decompression */
    while (cinfo.output_scanline < cinfo.output_height) {
	count = jpeg_read_scanlines(&cinfo, rowPtr, nLines);
        if (count <= 0) {
            Tcl_AppendResult(interp, "error reading jpeg file \"",
			     fileName, "\": ",
			     Tcl_Eof(chan) ? "not enough data" : Tcl_PosixError(interp),
			     (char *) NULL);
            ckfree((char *) pixelPtr);
	    ckfree((char *) rowPtr);
	    jpeg_destroy_decompress(&cinfo);
            return TCL_ERROR;
        }
        block.height = count;
	/* Ecriture dans la photo */
        Tk_PhotoPutBlock(imageHandle, &block, destX, destY, width, count);
        destY += count;
    }
    /* Fini, on libere les donnees */
    ckfree((char *) pixelPtr);
    ckfree((char *) rowPtr);
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    return TCL_OK;
}

/*----------------------------------------------------------------------*/

/**
 ** Lecture d'un morceau d'image
 **/
int
jpegDataReadProc(Tcl_Interp *interp, Tcl_Obj *data,
		 Tcl_Obj *formatString, Tk_PhotoHandle imageHandle,
		 int destX, int destY, int width, int height,
		 int srcX, int srcY)
{
    int fileWidth, fileHeight;
    int nLines, nBytes, h, type, count;
    unsigned char *pixelPtr;
    unsigned char **rowPtr;
    Tk_PhotoImageBlock block;
    
    /* Lecture de l'entete */
    type = dataReadJpegHeader(data, &fileWidth, &fileHeight);
    if (type == 0) {
	/* source non reconnue comme jpeg */
	Tcl_AppendResult(interp, 
			 "couldn't read jpeg header from data", NULL);
	return TCL_ERROR;
    }
    if ((fileWidth <= 0) || (fileHeight <= 0)) {
	/* Dimensions negatives ou nulles */
	Tcl_AppendResult(interp, "Jpeg data has dimension(s) <= 0", NULL);
	return TCL_ERROR;
    }
    /* Calcul de la zone a lire */
    if ((srcX + width) > fileWidth) {
        width = fileWidth - srcX;
    }
    if ((srcY + height) > fileHeight) {
        height = fileHeight - srcY;
    }
    if ((width <= 0) || (height <= 0)
        || (srcX >= fileWidth) || (srcY >= fileHeight)) {
        return TCL_OK;
    }
    /* Debut de la decompression */
    jpeg_start_decompress(&cinfo);
    
    if (type == TCL_JPEG_MONO) {
	block.pixelSize = 1;
	block.offset[0] = 0;
	block.offset[1] = 0;
	block.offset[2] = 0;
    } else {
	block.pixelSize = 3;
        block.offset[0] = 0;
	block.offset[1] = 1;
	block.offset[2] = 2;
    }
    block.width = width;
    block.pitch = block.pixelSize * fileWidth;

    /* Allocation structure photo */
    Tk_PhotoExpand(imageHandle, destX + width, destY + height);

#if 0
    /* XXX */
    if (srcY > 0) {
        Tcl_Seek(chan, (srcY * block.pitch), SEEK_CUR);
    }
#endif
    /*  buffer pour  le nombre de lignes prescrit par libjpeg */
    nLines = cinfo.rec_outbuf_height;
    nBytes = nLines * block.pitch;
    pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
    rowPtr = (unsigned char **) ckalloc((unsigned) nLines);
    /* Initialisation des pointeurs sur les lignes */
    for (h = 0; h < nLines; h++) {
	rowPtr[h] = pixelPtr + h*block.pitch;
    }
    block.pixelPtr = pixelPtr + srcX * block.pixelSize; /* XXXX ??? */

    /* Boucle de lecture/decompression */
    while (cinfo.output_scanline < cinfo.output_height) {
	count = jpeg_read_scanlines(&cinfo, rowPtr, nLines);
        if (count <= 0) {
            Tcl_AppendResult(interp, "error reading jpeg data ",
			     (char *) NULL);
            ckfree((char *) pixelPtr);
	    ckfree((char *) rowPtr);
	    jpeg_destroy_decompress(&cinfo);
            return TCL_ERROR;
        }
        block.height = count;
	/* Ecriture dans la photo */
        Tk_PhotoPutBlock(imageHandle, &block, destX, destY, width, count);
        destY += count;
    }
    /* Fini, on libere les donnees */
    ckfree((char *) pixelPtr);
    ckfree((char *) rowPtr);
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    return TCL_OK;
}


/*----------------------------------------------------------------------*/

/**
 ** Ecriture d'une image au format jpeg 
 **/
#ifdef notyet
int 
jpegFileWriteProc(Tcl_Interp *interp, const char *filename, 
		  Tcl_Obj *formatString,
		  Tk_PhotoImageBlock *blockPtr)
{
    return TCL_OK;
}
#endif

/*----------------------------------------------------------------------*/

/**
 ** Creation directe d'une photo depuis un channel au format JPEG 
 **/
int
JpegPhotoChannelRead(ClientData clientData, Tcl_Interp *interp,
		       int argc, char *argv[]) 
{
    Tk_PhotoHandle photoHandle;
    Tcl_Channel chan;
    int fileWidth, fileHeight;
    int nLines, nBytes, h, type, count;
    unsigned char *pixelPtr;
    unsigned char **rowPtr;
    Tk_PhotoImageBlock block;
    int chanMode;
    int destY = 0;

    if ((photoHandle = Tk_FindPhoto(interp, argv[1])) == NULL) {
	Tcl_AppendResult(interp, "image \"", argv[1], "\" doesn't",
			 " exist or is not a photo image", NULL);
	return TCL_ERROR;
    }
    if ((chan = Tcl_GetChannel(interp, argv[2], &chanMode)) == NULL) {
	Tcl_AppendResult(interp, "channel \"", argv[2], "\" is not",
			 " registered", NULL);
	return TCL_ERROR;
    }
    if (Tcl_SetChannelOption(interp, chan, "-translation", "binary")
	!= TCL_OK) {
	return TCL_ERROR;
    }
    /* Lecture de l'entete */
    type = fileReadJpegHeader(chan, &fileWidth, &fileHeight);
    if (type == 0) {
	/* source non reconnue comme jpeg */
	Tcl_AppendResult(interp, 
			 "couldn't read jpeg file header from channel", 
			 NULL);
	return TCL_ERROR;
    }
    if ((fileWidth <= 0) || (fileHeight <= 0)) {
	/* Dimensions negatives ou nulles */
	Tcl_AppendResult(interp, "Jpeg channel has dimension(s) <= 0", NULL);
	return TCL_ERROR;
    }
    Tk_PhotoSetSize(photoHandle, fileWidth, fileHeight);
    
    /* Debut de la decompression */
    jpeg_start_decompress(&cinfo);
    
    if (type == TCL_JPEG_MONO) {
	block.pixelSize = 1;
	block.offset[0] = 0;
	block.offset[1] = 0;
	block.offset[2] = 0;
    } else {
	block.pixelSize = 3;
        block.offset[0] = 0;
	block.offset[1] = 1;
	block.offset[2] = 2;
    }
    block.width = fileWidth;
    block.pitch = block.pixelSize * fileWidth;

    /*  buffer pour  le nombre de lignes prescrit par libjpeg */
    nLines = cinfo.rec_outbuf_height;
    nBytes = nLines * block.pitch;
    pixelPtr = (unsigned char *) ckalloc((unsigned) nBytes);
    rowPtr = (unsigned char **) ckalloc((unsigned) nLines);
    /* Initialisation des pointeurs sur les lignes */
    for (h = 0; h < nLines; h++) {
	rowPtr[h] = pixelPtr + h*block.pitch;
    }
    block.pixelPtr = pixelPtr + block.pixelSize; /* XXXX ??? */

    /* Boucle de lecture/decompression */
    while (cinfo.output_scanline < cinfo.output_height) {
	count = jpeg_read_scanlines(&cinfo, rowPtr, nLines);
        if (count <= 0) {
            Tcl_AppendResult(interp, "error reading jpeg channel: ",
			     Tcl_Eof(chan) ? "not enough data" :
			     Tcl_PosixError(interp),
			     (char *) NULL);
            ckfree((char *) pixelPtr);
	    ckfree((char *) rowPtr);
	    jpeg_destroy_decompress(&cinfo);
            return TCL_ERROR;
        }
        block.height = count;
	/* Ecriture dans la photo */
        Tk_PhotoPutBlock(photoHandle, &block, 0, destY, fileWidth, count);
        destY += count;
    }
    /* Fini, on libere les donnees */
    ckfree((char *) pixelPtr);
    ckfree((char *) rowPtr);
    jpeg_finish_decompress(&cinfo);
    jpeg_destroy_decompress(&cinfo);
    return TCL_OK;
}
		      
/*----------------------------------------------------------------------*/

/**
 ** Enregistre tout ce travail aupres de TCL 
 **/
void
registerJpegFormat(Tcl_Interp *interp)
{
    Tk_PhotoImageFormat jpegFormat;

    jpegFormat.name = "jpeg";
    jpegFormat.fileMatchProc = jpegFileMatchProc;
    jpegFormat.stringMatchProc = jpegDataMatchProc;
    jpegFormat.fileReadProc = jpegFileReadProc;
    jpegFormat.stringReadProc = jpegDataReadProc;
    jpegFormat.fileWriteProc = NULL;
    jpegFormat.stringWriteProc = NULL;

    Tk_CreatePhotoImageFormat(&jpegFormat);
    
    /* + une nouvelle commande */
    Tcl_CreateCommand(interp, "jpegread", JpegPhotoChannelRead, NULL,
		      NULL);
}

/*----------------------------------------------------------------------*/

int
Tkjpeg_Init(Tcl_Interp *interp)
{
    int code;

    code = Tcl_PkgProvide(interp, "Tkjpeg", "1.0");
    if (code != TCL_OK) {
	return code;
    }
    registerJpegFormat(interp);
    return TCL_OK;
}

/* C'est tout */
/*----------------------------------------------------------------------*/

