# 
# Copyright (C) 1996-2001 LAAS/CNRS 
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions
# are met:
# 
#    - Redistributions of source code must retain the above copyright
#      notice, this list of conditions and the following disclaimer.
#    - Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials provided
#      with the distribution.
# 
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
# COPYRIGHT HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
# 
# $LAAS: setup.tcl,v 1.31 2003/02/09 16:31:29 matthieu Exp $
#----------------------------------------------------------------------
package provide Gdhe::setup 1.0

global gdheNoControlPanel

###
### setup - mise en place de l'interface 
###
proc setup {} {
    global obsDist obsElev obsAzi obsX obsY obsZ
    global xmin ymin zmin xmax ymax zmax
    global polygon_mode perspective_mode
    global gdheBase gdheNoControlPanel

    # gdheBase is the widget in which the main window is
    # displayed. Defaults to ".".
    if {![info exists gdheBase]} { 
	set gdheBase ""
	wm title . "GDHE"
	wm geometry . +0+0
	wm deiconify .
    } elseif { "$gdheBase" != "" } {
	if {![winfo exists $gdheBase]} {
	    error "No such widget: $gdheBase"
	}
    }

#    frame .menuBar  -borderwidth 3 -relief raised
#    pack .menuBar -side top -fill x
#    puts "1"
#    menubutton .menuBar.file -text File -underline 0 -menu .menuBar.file.m
#    menu .menuBar.file.m -tearoff 0
#    .menuBar.file.m add command -label "Quit" -underline 0 -command exit
#    pack .menuBar.file -side left
#    menubutton .menuBar.help -text Help -underline 0 -menu .menuBar.help.m
#    menu .menuBar.help.m -tearoff 0
#    .menuBar.help.m add command -label "About..." -underline 0 \
#	-command aboutBox
#    pack .menuBar.help -side right

    # fenetre principale avec le widget OpenGL
    frame $gdheBase.gdhe
    
    set w $gdheBase.gdhe.ogl
    addWindow $w
    togl $w -width 640 -height 480 -privatecmap true -rgba true  \
	-double true -depth true -ident main
    setLights

    # geometrie des elements dans la fenetre
    pack $w  -side left -padx 3 -pady 3 -fill both -expand t
    pack $gdheBase.gdhe  -fill both -expand t

    # calcul de quelques valeurs utiles 
    set maxdist [expr hypot($ymax - $ymin, $xmax - $xmin)*2.0]

    # fenetre de controle 
    if {![info exists gdheNoControlPanel]} {
	toplevel $w.pos 
	wm title $w.pos "Position"
	# wm group $w.pos "$gdheBase."
	wm geometry $w.pos -0+0

	# Distance de l'observateur
	scale  $w.pos.dist  -length 200  -label {Distance} \
	    -from 0 -to $maxdist \
	    -resolution 0.1 \
	    -command "setSlider $w dist" -orient horizontal
	# Elevation
	scale  $w.pos.elev   -label {Elevation} -from -180 -to 180 \
	    -variable "::obsElev" \
	    -command "setSlider $w elev" -orient horizontal
	# Azimut
	scale $w.pos.azi -label {Azimut} -from -180 -to 180 \
	    -variable "::obsAzi" \
	    -command "setSlider $w azi" -orient horizontal
	# Position en X 
	scale $w.pos.x -label {Target X} -from $xmin -to $xmax \
	    -resolution 0.01\
	    -variable "::obsX" \
	    -command "setSlider $w x" -orient horizontal
	# Position en Y
	scale $w.pos.y -label {Target Y} -from $ymin -to $ymax \
	    -resolution 0.01\
	    -variable "::obsY" \
	    -command "setSlider $w y" -orient horizontal
	# Position en Z
	scale $w.pos.z -label {Target Z} -from $zmin -to $zmax \
	    -resolution 0.01\
	    -variable "::obsZ" \
	    -command "setSlider $w z" -orient horizontal

	label $w.pos.rate -text rate:

	frame $w.pos.props
	checkbutton $w.pos.props.light -text "Lights" \
	    -variable enable_lighting -command " $w redraw"

	set polygonModeMenu \
	    [tk_optionMenu $w.pos.props.fillStyle polygon_mode fill line]
	$w.pos.props.fillStyle configure -width 10
	$polygonModeMenu entryconfigure 0 -command "setPolygonMode $w"
	$polygonModeMenu entryconfigure 1 -command "setPolygonMode $w"
	set perpectiveModeMenu \
	    [ tk_optionMenu $w.pos.props.perspMode perspective_mode \
		  perspective orthogonal]
	$w.pos.props.perspMode configure -width 10
	$perpectiveModeMenu entryconfigure 0 \
	    -command "$w setObs -perspective ; $w redraw true"
	$perpectiveModeMenu entryconfigure 1 \
	    -command "$w setObs -ortho ; $w redraw true"
    

	grid $w.pos.props.light -column 0 -row 0 -columnspan 2
	grid $w.pos.props.fillStyle -column 0 -row 1
	grid $w.pos.props.perspMode -column 1 -row 1

	# Boutton Print
	button $w.pos.print -text "Dump EPS" -command "print $w"

	# Boutton Quit
	button $w.pos.btn  -text Quit -command exit

	# Geometry managment
	pack $w.pos.dist  -fill x
	pack $w.pos.elev  -fill x
	pack $w.pos.azi -fill x
	pack $w.pos.x -fill x
	pack $w.pos.y -fill x
	pack $w.pos.z -fill x
	pack $w.pos.rate -fill x
	pack $w.pos.props -fill x
	pack $w.pos.print -fill x
	pack $w.pos.btn -fill x

	#  position initiale des sliders
	$w.pos.dist set $obsDist
	$w.pos.elev set $obsElev
	$w.pos.azi set $obsAzi
	$w.pos.x set $obsX
	$w.pos.y set $obsY
	$w.pos.z set $obsZ
    }

    # Handler des evenements souris dans la fenetre principale
    bind $w <ButtonPress-1> {
	set oldX %x
	set oldY %y
    }
    bind $w <ButtonPress-2> {
	set oldX %x
	set oldY %y
    }
    bind $w <ButtonPress-3> {
	set oldX %x
	set oldY %y
    }
    #     bind $w <B3-Double> {
    # 	# maj sliders
    # 	if [winfo exists $w.pos] {
    # 	    $w.pos.x set 0
    # 	    $w.pos.y set 0
    # 	}
    # 	# maj vue 
    # 	$w setObs -x 0 -y 0
    # 	$w redraw
    #     }
    bind $w <B1-Motion> [concat "motion_event_b1 $w" { 
	[lindex [%W config -width] 4] \
	    [lindex [%W config -height] 4] \
	    [expr %x - $oldX] [expr %y - $oldY]
	set oldX %x
	set oldY %y
    }]
    bind $w <B2-Motion> [concat "motion_event_b2 $w $maxdist" {
	[lindex [%W config -width] 4] \
		[lindex [%W config -height] 4] \
		[expr %x - $oldX] [expr %y - $oldY]
	set oldX %x
	set oldY %y
    }]
    bind $w <B3-Motion> [concat "motion_event_b3 $w" {
	[lindex [%W config -width] 4] \
		[lindex [%W config -height] 4] \
		[expr %x - $oldX] [expr %y - $oldY]
	set oldX %x
	set oldY %y
    }]
    focus $w
    bind $w <Key-Up> "move $w z +"
    bind $w <Key-Down> "move $w z -"
    bind $w <Key-Left> "move $w x -"
    bind $w <Key-Right> "move $w x +"
    bind $w <Shift-Key-Left> "move $w y -"
    bind $w <Shift-Key-Right> "move $w y +"

    if {$gdheBase == ""} {
	menu .windowMenu -tearoff 0
	.windowMenu add command -label "Quit" -command exit
    }

    #    bind $w <ButtonPress-3> { tk_popup .windowMenu %X %Y }

    # chargement des modeles 
    load_env $w;
    
    $w setObs -dist $obsDist -elev $obsElev -azi $obsAzi \
	    -x $obsX -y $obsY -z $obsZ
    $w redraw
}


#----------------------------------------------------------------------

##
## Deplacement du point de vue
##
proc move {w axis sign} {
    upvar \#0 obs[string toupper $axis] pos

    set pos [expr $pos $sign 0.1]
    $w setObs -$axis $pos
    $w redraw
}
	      

#----------------------------------------------------------------------

##
## draw_all - Fonction d'affichage de la scene
##
## appelee par Togl chaque fois qu'il faut dessiner une fenetre
##


proc draw_all {w} {
    global robots pos platform load trajectory rs_trajectory walls \
	window displayTime transp
    set window $w
    # puts "draw_all $w"
    draw_env 
    draw_per_obst
    # dessin des robots
    foreach {r} [array names robots] {
	if {[info exists pos($r)]} {
	    pushMatrix
	    switch [llength $pos($r)] {
		# position a trois parametres x y theta
		3 {
		    set x [lindex $pos($r) 0]
		    set y [lindex $pos($r) 1]
		    set theta [lindex $pos($r) 2]
		    translate $x $y 0
		    rotate $theta 0 0 1
		}
		# position a six parametres: angles de bryant
		6 {
		    set cap [lindex $pos($r) 0]
		    set roulis [lindex $pos($r) 1]
		    set tangage [lindex $pos($r) 2]
		    set x [lindex $pos($r) 3]
		    set y [lindex $pos($r) 4]
		    set z [lindex $pos($r) 5]
		    translate $x $y $z
		    rotate $cap 0 0 1
		    rotate $roulis 0 1 0
		    rotate $tangage 1 0 0 
		}
		default {
		    puts "bizarre llength pos != 3 ou 6"
		}
	    }
	    eval $robots($r)
	    if {[info exists platform($r)]} {
		eval $platform($r)
	    }
	    if {[info exists load($r)]} {
		eval $load($r)
	    }
	}
	popMatrix

	# dessin des murs attaches au robot
	draw_walls $r

	# dessin le RS_trajectoire du robot
	if {[info exists rs_trajectory($r)]} {
	    draw_rs_trajectory $r
	}
	if {[info exists trajectory($r)]} {
	    draw_trajectory $r
	}

    }
    # dessin des parties transparentes
    foreach {r} [array names transp] {
	if {[info exists pos($r)]} {
	    pushMatrix
	    switch [llength $pos($r)] {
		# position a trois parametres x y theta
		3 {
		    set x [lindex $pos($r) 0]
		    set y [lindex $pos($r) 1]
		    set theta [lindex $pos($r) 2]
		    translate $x $y 0
		    rotate $theta 0 0 1
		}
		# position a six parametres: angles de bryant
		6 {
		    set cap [lindex $pos($r) 0]
		    set roulis [lindex $pos($r) 1]
		    set tangage [lindex $pos($r) 2]
		    set x [lindex $pos($r) 3]
		    set y [lindex $pos($r) 4]
		    set z [lindex $pos($r) 5]
		    translate $x $y $z
		    rotate $cap 0 0 1
		    rotate $roulis 0 1 0
		    rotate $tangage 1 0 0 
		}
		default {
		    puts "bizarre llength pos != 3 ou 6"
		}
	    }
	    eval $transp($r)
	}
	popMatrix
    }
    # Affichage eventuel du taux de rafraichissement 
    if {[winfo exists $w.pos.rate]} {
	if {[info exists displayTime]} {
	    if { $displayTime > 10 } {
		set rate [format "%.1f" [expr 1000.0/$displayTime]]
		$w.pos.rate configure -text "rate: $rate frames/s"
	    } else {
		$w.pos.rate configure -text "display time $displayTime ms"
	    }
	} else {
	    $w.pos.rate configure -text ""
	}
    }
}

#----------------------------------------------------------------------
#
# Essai de seconde fenetre d'affichage
#
proc aux_window {} {

    if ![winfo exists .aux] {
	toplevel .aux
	togl .aux.ogl -width 320 -height 200 -rgba true -double true \
	    -depth true -ident aux -sharelist main
	pack .aux.ogl  -side left -padx 3 -pady 3 -fill both -expand t 
	
	addWindow .aux.ogl

	.aux.ogl setObs -dist 4 -elev 90 -azi 0 -x 0 -y 0 -z 0 
	.aux.ogl setObs -ortho
	.aux.ogl redraw
	setLights
    }
}

#----------------------------------------------------------------------

## 
## Procedure de mise a jour d'une fenetre de camera montee sur un robot
##
proc update_robot_window { name index op } {
    upvar $name pos

    # nom de la fenetre 
    set w .on$index.ogl
    if ![winfo exists $w] {
	# rien a faire si la fenetre n'existe pas
	return {}
    }
    # recupere les coordonnees 
    # Pour l'instant on ne traire que les robots a 3 coordonees 
    set x [ lindex $pos($index) 0]
    set y [ lindex $pos($index) 1]
    set theta [lindex $pos($index) 2]
    # positionne l'observateur
    $w setObs -dist 1 -elev 20 -azi [expr 90-$theta] -x $x -y $y -z 1.3
    # reaffiche 
    $w redraw
}

#----------------------------------------------------------------------

##
## Creation d'une camera montee sur un robot
##
proc robot_window {r} {
    global pos

    set w .on$r

    if ![winfo exists $w] {
	toplevel $w
	wm title $w "Camera on $r"

	togl $w.ogl -width 640 -height 480 -rgba true -double true \
	    -depth true -ident $r -sharelist main
	pack $w.ogl  -side left -padx 3 -pady 3 -fill both -expand t 
	
	addWindow $w.ogl
	update_robot_window pos $r r

	setLights
    }
    # Trace les modifications de la position du robot
    trace variable pos r update_robot_window 
}

#----------------------------------------------------------------------

proc setPolygonMode {w} {
    global polygon_mode

    $w polygonMode $polygon_mode
    $w redraw t
}
#----------------------------------------------------------------------

##
## motion_event - Handler des evenements dans la fenetre OpenGL
##    le bouton gauche modifie l'azimut (deplacements horizontaux) 
##    et l'elevation (deplacements verticaux) de l'observateur
##    le bouton du milieu modifie la distance au point vise
##    le bouton droit modifie le x,y du point vise
# cas du bouton de gauche
proc motion_event_b1 { w width height dx dy } { 
    global obsElev obsAzi

    # Calcul nouveaux azi et elev
    set elev [expr $obsElev + 180.0 * $dy / $height ]
    if { $elev > 180 } { set elev [expr $elev - 360] }
    if { $elev < -180 } { set elev [expr $elev + 360] }
    set obsElev $elev

    set azi [expr $obsAzi + 360.0 * $dx/$width ]
    if { $azi > 180 } { set azi [expr $azi - 360] }
    if { $azi < -180 } { set azi [expr $azi + 360] }

    # maj sliders
    set obsAzi $azi

    # maj vue 
    $w setObs -elev  $obsElev -azi $obsAzi 
    $w redraw
}

proc motion_event_b2 { w maxdist width height dx dy } { 
    global obsDist

    # Calcul nouvelle distance
    set dist [expr $obsDist + $maxdist * $dy / $height ]
    if { $dist > $maxdist } { set dist $maxdist }
    if { $dist < 0 } { set dist 0 }

    # maj sliders
    set obsDist $dist

    # maj vue 
    $w setObs -dist  $obsDist
    $w redraw
}

proc motion_event_b3 { w width height dx dy } { 
    global obsElev obsAzi obsX obsY
    global xmin ymin zmin xmax ymax zmax

    # Calcul nouvelle distance
    set azimuth [expr $obsAzi * 3.1415926 / 180]
    set elev [expr $obsElev * 3.1415926 / 180]

    set X [expr $obsX - ($xmax - $xmin) * $dx * cos( $azimuth) / $width ]
    set Y [expr $obsY + ($xmax - $xmin) * $dx * sin( $azimuth) / $width ]

    set X [expr $X  + ($ymax - $ymin) * $dy * sin( $azimuth) / $height ]
    set Y [expr $Y  + ($ymax - $ymin) * $dy * cos( $azimuth) / $height ]

#    set obsX [expr $dy * cos($azimuth)]
#     set DX [ expr  $dx * cos ($azimuth) - $dy * sin($azimuth) ]
#     set DY [ expr ( - $dx * sin ($azimuth ) - $dy * cos($azimuth) )]
#     set X [ expr $DX * cos($elev) + $DY * sin($elev)]
#     set Y [ expr -$DX * sin($elev) + $DY*cos($elev)]
#     set obsX [expr [$w.pos.x get] + ($xmax - $xmin) * $X / $width ]
#     set obsY [expr [$w.pos.y get] + ($ymax - $ymin) * $Y / $height ]

    if { $X > $xmax } { set X $xmax }
    if { $X < $xmin } { set X $xmin }
    if { $Y > $ymax } { set Y $ymax }
    if { $Y < $ymin } { set Y $ymin }

    # maj sliders
    set obsX $X
    set obsY $Y

    # maj vue 
    $w setObs -x $obsX -y $obsY
    $w redraw
}

#----------------------------------------------------------------------

##
## setSlider - Handler des evenements sur les sliders du panneau de controle
##
proc setSlider {w axis value} {

    # Mise a jour de la variable qui a change
    switch -exact $axis {
	dist { $w setObs -dist $value }
	elev { $w setObs -elev $value }
	azi { $w setObs -azi $value }
	x { $w setObs -x $value }
	y { $w setObs -y $value }
	z { $w setObs -z $value }

    }
    # reaffichage 
    $w redraw
}

#----------------------------------------------------------------------
###
### chargement des autres fichiers
###
proc load_env {w} {
    global gdhe_library
    global station_vertices station_boundaries 
    global cell_vertices cell_boundaries
    global obstacle_vertices obstacle_boundaries
    global obsDist obsElev obsAzi obsX obsY obsZ
    global envDefined
    
    # modeles d'objets
    catch {
	package require Gdhe::models
    } result

    Gdhe::models::init

    # gestion de l'environnement 
    package require Gdhe::env

    # eclairage de la scene
    package require Gdhe::lights
    Gdhe::lights::setLights
}

#----------------------------------------------------------------------
###
### definition d'un 'objet'
###
### Cette fonction utilise la variable globale objects 
### qui indique si un objet est defini et quel est l'indice de la 
### display list associee
###
proc object {name def} {
    global objects

    # Teste si cet objet a deja ete defini
    if {![info exists objects($name)]} {
	# Alloue un indice de display list
	set objects($name) [genLists 1]
	# Memorise les traces suivants dans la display list
	newList $objects($name)
	# Evalue la definition de l'objet dans le contexte appelant
	uplevel 1 $def
	# Fin de la definition de la display list
	endList 
    } else {
	# L'objet est deja connu - appelle la display list correspondante
	callList $objects($name)
    }
}

#----------------------------------------------------------------------
##
## Destruction d'un objet
##
proc objectDelete {name} {
    global objects

    if { [info exists objects($name)]} {
	# Destruction de la display list associe
	deleteLists $objects($name)
	# detruit la reference dans le tableau global objects
	unset objects($name)
    }
}
#----------------------------------------------------------------------
###
### Gestion de la liste des fenetres 
###

#----------------------------------------------------------------------
##
## Ajout d'une fenetre a la liste
##
proc addWindow {w} {
    global allWindows

    if {[info exists allWindows]} {
	set allWindows [concat $allWindows $w]
    } else {
	set allWindows $w
    }
}

#----------------------------------------------------------------------
##
## Reaffichage de toutes les fenetres
##
## f: true force un reaffichage immediat
##    sinon reaffichage lorsque inactif
##
proc redrawAllWindows {{f false}} {
    global allWindows

    foreach w $allWindows {
	$w redraw $f
    }    
}

#----------------------------------------------------------------------

proc aboutBox {} {
 tk_messageBox -icon info -type ok -title "About GDHE" \
 -message \
     {Graphic Display For Hilare Experiments
Author: Matthieu Herrb

Copyright (c) 1997-1999 LAAS/CNRS}
}

#----------------------------------------------------------------------
##
## Dialogue d'impression
##
proc print {w} {
    set types {
	{{EPS Files} {.eps}}
	{{All Files} *}
    }
    set filename [tk_getSaveFile -title "Save EPS file" -filetypes $types]
    if {$filename != ""} {
	$w dumpEps -color $filename
    }
}

#----------------------------------------------------------------------
###
### Procedure appelee pour autoriser un client a se connecter
###
proc clientAuthorize {ip name} {
    global allowedClients deniedClients

    foreach c $allowedClients {
	if {[regexp $c $ip] || [regexp $c $name]} {
	    return 
	}
    }
    foreach c $deniedClients {
	if {[regexp $c $ip] || [regexp $c $name]} {
	    error "Connection refused from $ip"
	}
    }
    error "Connection refused from $ip"
}

proc allow {exp} {
    global allowedClients
    
    lappend allowedClients $exp
}

proc deny {exp} {
    global deniedClients

    lappend deniedClients $exp
}

#----------------------------------------------------------------------
###
### code execute' au demarrage 
###


# valeurs par defaut
set obsDist 3
set obsElev 0
set obsAzi 0
set obsX 0
set obsY 0
set obsZ 0
# dimensions
set xmin -10.0
set xmax 10.0
set ymin -10.0
set ymax 10.0
set zmin -5.0
set zmax 5.0
  
set polygonMode fill  

#----------------------------------------------------------------------
#  Access control

allow "^127.0.0.1"
allow "^localhost$"

allow "\.laas\.fr$"
allow "^140\.93\."
allow "^195\.83\.132\."
allow "^192\.168\."
deny ".*"

